#ifndef AMBAARBITER_H_
#define AMBAARBITER_H_

#include "systemc.h"
#include "enum_master_sig.h"
#include "amba_arbiter_if.h"
#include "amba_master.h"
#include "amba_bus.h"
#include "amba_request.h"
#include "amba_data.h"

using namespace master;
/**
 * @author Andre Silva
 * @author Guilherme Alvaro
 * 
 * @brief bus Namespace
 */
namespace bus{
	
	 /**
	 * @brief Class Amba Arbiter that implements the Fix Priority Algorithm
	 * @author Andre Silva
	 * @author Guilherme Alvaro
	 */
	 
	class Amba_Arbiter_Fix_Priority: public Amba_Arbiter_if, public sc_module {
		
		private: 

		/**
		 * @brief Pointer to array of AmbaMaster that are in split state
		 */
			AmbaMaster** split_master;
			
		/**
		 * @brief Pointer to master default
		 */
			AmbaMaster* master_default;

		/**
		 * @brief Pointer to the actual master of bus
		 */

			AmbaMaster* actual_master;

		/**
		 * @brief Auxiliar pointer used to define the master of bus
		 */

			AmbaMaster* temp_master;

		/**
		 * @brief Pointer to the bus
		 */
			AmbaBus* amba_bus;
			
		/**
		 * @brief Type of the actual request
		 */
			
			Amba_Request_Type actual_request_type;

		/**
		 * @brief Pointer to clock system
		 */
			sc_clock* clock;

		/**
		 * @brief Event that notifies the release of bus
		 */
			sc_event release_bus;	

		/**
		 * @brief Value of highest priority of the masters that request the bus
		 */
			char highest_priority;

		/**
		 * @brief Indicates if the bus is busy
		 */
			bool busy_bus;

		/**
		 * @brief Indicates if there are any master using the bus
		 */
			bool exist_master_allocated;

		/**
		 * @brief Create a request. Used by master default
		 * @param AmbaRequest& req : Request package
		 */
			void mount_request(AmbaRequest & req);
			
		/**
		 * @brief Verifies if the master passed is in split state
		 * @param AmbaMaster& master:  device master
		 * @return If the master passed is in split state
		 */
			bool master_is_split(AmbaMaster& master);

		public:
		
			SC_HAS_PROCESS(Amba_Arbiter_Fix_Priority);
	
			/**
			 *  Constructor of the Class Amba_Arbiter_Fix_Priority
			 * @param sc_module_name nm Name of module
			 * @param sc_clock* clk Point for the Clock System
			 */
			Amba_Arbiter_Fix_Priority(sc_module_name nm, sc_clock* clk);
	

			/**
			 * @brief Requests the bus for the AmbaMaster passed
			 * @param AmbaMaster& master the master device
			 * @param Amba_Request_Type req Type of Request
			 * @return bool if the AmbaMaster get the bus
			 */
			bool alloc_master(AmbaMaster& master, Amba_Request_Type req);
			
			/**
			 * @brief Verifies if the AmbaMaster is actual master of the bus
			 * @param AmbaMaster& master the master device
			 * @return bool if the AmbaMaster is actual master of the bus
			 */
			bool isMaster(AmbaMaster& master);
			

			/**
			 * @brief Returns  the actual master of the bus
			 * @return int The ID of the actual master of the bus
			 */	
			int ownerBus();
			
			/**
			 * @brief Returns the type of the actual transfer
			 * @returns bool the type of the actual transfer
			 */	
			bool locked_transfer();
			
			/**
			 * @brief Notifies the arbiter about the end of transfer
			 */	
			void notify_release_bus();
			
			/**
			 * @brief Verifies if the bus is busy
			 * @return bool The actual state of the bus
			 */	
			bool is_busy_bus();
			
			/**
			 * @brief Sets the bus managed by arbiter
			 * @param AmbaBus* amba_bus Pointer for the new AmbaBus
			 */	
			void setBus(AmbaBus* amba_bus);
			
		/**
		 * @brief Thread that simulates the behaviour of the master default
		 */
			void arbiter_thread();		
	

			/**
			 * @brief Changes the actual state of bus for busy 
			 */	
			void turn_busy_bus();


			/**
			 * @brief sets the locked_transfer internal variable
			 * @param req New value for the locked_transfer variable
			 */
			void set_locked_transfer(Amba_Request_Type req);
			

			/**
			 * @brief Inits one split transaction
			 * @param AmbaMaster& master The actual master of the bus
			 * @param AmbaSlave& slave The Slave that requests the Split transaction
			 */	
			void init_split_transaction(AmbaMaster& master, AmbaSlave& slave);


			/**
			 * @brief Ends one split transaction
			 * @param AmbaSlave& slave The Slave that initiated the Split transaction
			 */	
			void end_split_transaction(AmbaSlave& slave);
			

			/**
			 * @brief Returns the actual Master of the bus
			 * @return AmbaMaster* Pointer for the actual master of the bus
			 */	
			AmbaMaster* get_actual_master();


			/**
			 * @brief Returns if the slave passed is in a split transaction
			 * @return bool if the slave passed is in a split transaction
			 */	
			bool slave_in_split_transaction(AmbaSlave& slave);
	};
}
#endif /*AMBAARBITER_H_*/
