#include "amba_bus.h"

/*********************************************************/
/*	      Master Methods								 */
/*														 */
/*********************************************************/

/**
 * Request the bus. This method is non-blocking. 
 */ 

bool AmbaBus::master_request_bus_non_blocking(AmbaMaster& master,  Amba_Request_Type req) {

	wait(clock->posedge_event()); 
	bool isAlloc = arbiter->alloc_master(master, req);

	wait(clock->posedge_event()); 
	 
	if (isAlloc)
		pipeline_mode = false;
	
	return isAlloc;
}
			
/**
 * Request the bus. This method is blocking.
 */
			 
void AmbaBus::master_request_bus_blocking(AmbaMaster& master,  Amba_Request_Type req){
	bool isAlloc = false;
	while (!isAlloc)
		isAlloc = master_request_bus_non_blocking(master, req);
}
			
						
/**
 * Send the controls signals and adress.
 */
 
bool AmbaBus::master_send_controls_signals(AmbaMaster& master, AmbaRequest& req){
	if(arbiter->isMaster(master)){
		decoder->decoder_address(req);
		this->amba_request = req;
		if (!pipeline_mode){
			wait(SC_ZERO_TIME);
			control_signals_sended = true;
		} else{
			this->control_signals_event.notify();
		}
		arbiter->turn_busy_bus();
		return true;
	} else return false;
}


/**
 * Send data.
 */			
			 
void AmbaBus::master_send_data(AmbaMaster& master, AmbaData& data, AmbaResponse& ambaResponse){

	if(arbiter->isMaster(master)){
		if (!pipeline_mode){
			wait(this->control_signals_event);
			pipeline_mode = true;		
		}

		this->amba_data = data;
		wait(SC_ZERO_TIME);
		alocated_data = true;
		wait(this->alocated_response_event);
		
		ambaResponse.HREADY = this->amba_response.HREADY;
		ambaResponse.HRESP  = this->amba_response.HRESP;

	}else{
		ambaResponse.HREADY = false;
		ambaResponse.HRESP = ERR;
	}
}


/**
 * Read data.
 */			
 
void AmbaBus::master_read_data(AmbaMaster& master, AmbaData& data, AmbaResponse& ambaResponse){
	if(arbiter->isMaster(master)){
		
		if (!pipeline_mode){
			wait(this->control_signals_event);
			pipeline_mode = true;		
		}

		wait(alocated_response_event & alocated_data_event);

		ambaResponse.HREADY = this->amba_response.HREADY;
		ambaResponse.HRESP  = this->amba_response.HRESP;

		data = this->amba_data;

	}else{
		ambaResponse.HREADY = false;
		ambaResponse.HRESP = ERR;
	}
}
		 
			 
/*********************************************************/
/*	      Slave Methods								 	 */
/*														 */
/*********************************************************/
			
void AmbaBus::slave_receive_controls_signals(AmbaSlave& slave, AmbaRequest& ambaRequest){

	while(true){
		wait(control_signals_event);
		if(decoder->get_actual_slave() == slave.getId()){
			break;
		}
	}
	ambaRequest = this->amba_request;
	if (ambaRequest.HTRANS == IDLE){
		wait(SC_ZERO_TIME);
		release_bus = true;
	}
}
			
void AmbaBus::slave_read_data(AmbaSlave& slave, AmbaResponse& response, AmbaData& data){
	if(decoder->get_actual_slave() == slave.getId()){
		this->amba_response.HREADY = response.HREADY;
		this->amba_response.HRESP = response.HRESP;
		if (response.HRESP == SPLIT){
			arbiter->init_split_transaction((*arbiter->get_actual_master()), slave);
		}
		wait(SC_ZERO_TIME);
		alocated_response = true;
		if (response.HREADY == true)
			release_bus = true;
		wait(alocated_data_event);
		data = amba_data;
	}
}
			
			
void AmbaBus::slave_send_data(AmbaSlave& slave, AmbaResponse& response, AmbaData& data){
	if(decoder->get_actual_slave() == slave.getId()){
		this->amba_response.HREADY = response.HREADY;
		this->amba_response.HRESP = response.HRESP;
		if (response.HRESP == SPLIT)
			arbiter->init_split_transaction((*arbiter->get_actual_master()), slave);
		wait(SC_ZERO_TIME);
		alocated_response = true;
		if (response.HREADY = true)
			release_bus = true;
		this->amba_data = data;
		alocated_data = true;
	}
}
	
void AmbaBus::check_sign_thread(){
	while(true){
		if(release_bus){
			if(check_release_bus()){
				arbiter->notify_release_bus();
				release_bus = false;
			}
		}
	
		if(control_signals_sended){
			this->control_signals_event.notify();
			this->control_signals_sended = false;
		}
		
		if(alocated_data){
			
			this->alocated_data_event.notify();
			this->alocated_data = false;
			
		}
		
		if(alocated_response){
			this->alocated_response_event.notify();
			this->alocated_response = false;
		}
		
		wait();
	}
	
}  
	
bool AmbaBus::check_release_bus(){
		
	if((this->amba_response.HREADY == true)
	   && (!arbiter->locked_transfer()))
		   return true;

	return false;
}

void AmbaBus::reset_signals(){

	release_bus = false;
	control_signals_sended = false;
	alocated_data = false;
	alocated_response = false;
	amba_response.HREADY = true;
	pipeline_mode = false;
	
}

AmbaBus::AmbaBus(sc_module_name nm, Amba_Arbiter_if* arb, sc_clock* clk): 
		sc_channel(nm),
		arbiter(arb), 
		clock(clk) {
			//here
			initMasterArray();
			
			reset_signals();
			SC_THREAD(check_sign_thread);
			   sensitive << clock->posedge_event();

			this->decoder = new AmbaDecoder("decoder");
			(this->arbiter)->setBus(this);
			(this->decoder)->setBus(this);
}


bool AmbaBus::is_master(AmbaMaster& master){
	return arbiter->isMaster(master);
}

void AmbaBus::master_release_bus(AmbaMaster& master){
	if (arbiter->isMaster(master)){
		arbiter->set_locked_transfer(UNLOCKED);
		release_bus = true;
	}
}
	
void AmbaBus::slave_end_split(AmbaSlave& slave){
	if (arbiter->slave_in_split_transaction(slave))
		arbiter->end_split_transaction(slave);
}

void AmbaBus::initMasterArray(){

	for(int i=0; i < 17; i++)
	{
		masterArray[i] = 0;
	}

}

void AmbaBus::register_master(AmbaMaster* master){
	for(int i = 0; i < 16; i ++){
		if (!masterArray[i]){
		 	master->setId(i);
			masterArray[i] = 1;
			break;
		}
		if (i == 16){
			cout << "Numero de Mestres acima do permitido \n";
			exit(-1);
		}
	}
}
	
void AmbaBus::register_slave(AmbaSlave* slave){
	this->decoder->registerSlave(slave);
}
