#ifndef AMBABUS_H_
#define AMBABUS_H_

#include "systemc.h" 
#include "amba_request.h"
#include "amba_response.h"
#include "amba_master.h"
#include "amba_slave.h"
#include "amba_arbiter_if.h"
#include "amba_master_tl1_if.h"
#include "amba_slave_tl1_if.h"

#include "enum_slave_sig.h"
#include "amba_data.h"
#include "amba_response.h"
#include "amba_decoder.h" 


using namespace slave;
using namespace amba;


class AmbaDecoder;
class Amba_Arbiter_if;

/**
 * @brief Class that represents the Amba Bus
 * @author Andre Silva
 * @author Guilherme Alvaro
 */

class AmbaBus : 
public sc_channel, 
public Amba_Master_tl1_if, 
public Amba_Slave_tl1_if
{

private: 


	/**
	 * @brief Pointer to the arbiter
	 */
	Amba_Arbiter_if* arbiter;

	/**
	 * @brief Pointer to the clock system
	 */
	sc_clock* clock;

	/**
	 * @brief Pointer to the decoder
	 */
	AmbaDecoder* decoder;

	/**
	 * @brief Data transfered by bus
	 */
	AmbaData amba_data;

	/**
	 * @brief Request transfered by bus
	 */
	AmbaRequest amba_request;

	/**
	 * @brief Response transfered by bus
	 */
	AmbaResponse amba_response;

	/**
	 * @brief Indicates the pipeline mode operation
	 */
	bool pipeline_mode;

	/**
	 * @brief Indicates the release of the bus
	 */
	bool release_bus;

	/**
	 * @brief Indicates the controls signals was posted in the bus
	 */
	bool control_signals_sended;

	/**
	 * @brief Indicates the data was sended to the bus
	 */
	bool alocated_data;

	/**
	 * @brief Indicates the response was sended to the bus
	 */
	bool alocated_response;


	/**
	 * @brief Event associated the coming of the control signals
	 */


	sc_event control_signals_event;

	/**
	 * @brief Event associated the coming of the data
	 */


	sc_event alocated_data_event;

	/**
	 * @brief Event associated the coming of the response
	 */

	sc_event alocated_response_event;
	
	
	/**
	 * @brief Verifies if any data, request or response was posted in the bus
	 */

	void check_sign_thread();	

	/**
	 * @brief Verify if the bus was released
	 * @return boool if the bus was released
	 */

	bool check_release_bus();	

	/**
	 * @brief Reset the signals of the bus
	 */

	void reset_signals();
	
	//here

	/**
	 * @brief Indicates the positions used in the bus
	 */

	int masterArray[17];	

	/**
	 * @brief Initializes the masterArray
	 */

	void initMasterArray();
	

 
	
public:

	SC_HAS_PROCESS(AmbaBus);
	
	/**
	 * @brief Constructor of class AmbaBus
	 * @param sc_module_name nm: Name of Bus
	 * @param Amba_Arbiter_if* arb: Pointer to the arbiter
	 * @param sc_clock* clk: pointer to the clock system
	 */
	AmbaBus(sc_module_name nm, Amba_Arbiter_if* arb, sc_clock* clk); 

	/*********************************************************/
	/*	      Master Methods								 */
	/*														 */
	/*********************************************************/
	
	bool master_request_bus_non_blocking(AmbaMaster& master, Amba_Request_Type req);			

	void master_request_bus_blocking(AmbaMaster& master, Amba_Request_Type req);
	
	bool master_send_controls_signals(AmbaMaster& master, AmbaRequest& req);
	
 	void master_send_data(AmbaMaster& master, AmbaData& data, AmbaResponse& ambaResponse);			 

    void master_read_data(AmbaMaster& master, AmbaData& data, AmbaResponse& ambaResponse);

	void master_release_bus(AmbaMaster& master);

	void slave_receive_controls_signals(AmbaSlave& slave, AmbaRequest& ambaRequest);			

	void slave_read_data(AmbaSlave& slave, AmbaResponse& response, AmbaData& data);
	
	void slave_send_data(AmbaSlave& slave, AmbaResponse& response, AmbaData& data);

	void slave_end_split(AmbaSlave& slave);
	
	bool is_master(AmbaMaster& master);
	
	void register_master(AmbaMaster* master);
	
	void register_slave(AmbaSlave* slave);
};
#endif /*AMBABUS_H_*/
